@@ -1,6 +1,7 @@
 from .capture import Capture  # noqa
 from .capture import CaptureResponse  # noqa
 from .capture import SetCookieParam, Cookie  # noqa
+from .capture import TrustedTimestampSettings # noqa
 from .helpers import get_devices  # noqa
 from .exceptions import (PlaywrightCaptureException, UnknownPlaywrightDeviceType,  # noqa
                          UnknownPlaywrightBrowser, UnknownPlaywrightDevice,
@@ -9,6 +10,7 @@ from .exceptions import (PlaywrightCaptureException, UnknownPlaywrightDeviceType
 __all__ = [
     'Capture',
     'CaptureResponse',
+    'TrustedTimestampSettings',
     'SetCookieParam', 'Cookie',
     'get_devices',
     'PlaywrightCaptureException',
