@@ -4,7 +4,7 @@ from __future__ import annotations
 
 import asyncio
 import binascii
-import json
+# import hashlib
 import logging
 import os
 import random
@@ -12,7 +12,7 @@ import re
 import sys
 import time
 
-from base64 import b64decode
+from base64 import b64decode, b64encode
 from io import BytesIO
 from logging import LoggerAdapter, Logger
 from tempfile import NamedTemporaryFile
@@ -23,6 +23,7 @@ from zipfile import ZipFile
 
 import aiohttp
 import dateparser
+import orjson
 
 from aiohttp_socks import ProxyConnector
 from bs4 import BeautifulSoup
@@ -32,6 +33,7 @@ from playwright.async_api import async_playwright, Frame, Error, Page, Download,
 from playwright.async_api import TimeoutError as PlaywrightTimeoutError
 from playwright_stealth import Stealth, ALL_EVASIONS_DISABLED_KWARGS  # type: ignore[attr-defined]
 from puremagic import PureError, from_string
+from rfc3161_client import TimestampRequestBuilder, TimeStampRequest, HashAlgorithm
 from w3lib.html import strip_html5_whitespace
 from w3lib.url import canonicalize_url, safe_url_string
 
@@ -119,6 +121,10 @@ class CaptureResponse(TypedDict, total=False):
     downloaded_file: bytes | None
     children: list[CaptureResponse] | None
 
+    # if the capture is triggered with with_trusted_timestamps, the response contains a
+    # dict[<entry name>] = '<base64 encoded timestamp response>'
+    trusted_timestamps: dict[str, str] | None
+
     # One day, playwright will support getting the favicon from the capture itself
     # favicon: Optional[bytes]
     # in the meantime, we need a workaround: https://github.com/Lookyloo/PlaywrightCapture/issues/45
@@ -135,6 +141,13 @@ class PlaywrightCaptureLogAdapter(LoggerAdapter):  # type: ignore[type-arg]
         return msg, kwargs
 
 
+class TrustedTimestampSettings(TypedDict, total=False):
+
+    url: str
+    hash_algorithm: HashAlgorithm | None
+    # NOTE: can add other settings such as auth mechanism, if needed.
+
+
 # good test pages:
 # https://kaliiiiiiiiii.github.io/brotector/?crash=false
 # https://www.browserscan.net/bot-detection
@@ -155,7 +168,7 @@ class Capture():
                  socks5_dns_resolver: str | list[str] | None=None,
                  general_timeout_in_sec: int | None=None, loglevel: str | int='INFO',
                  uuid: str | None=None, headless: bool=True,
-                 *, init_script: str | None=None):
+                 *, init_script: str | None=None, tt_settings: TrustedTimestampSettings | None=None):
         """Captures a page with Playwright.
 
         :param browser: The browser to use for the capture.
@@ -219,6 +232,8 @@ class Capture():
 
         self._init_script = init_script
 
+        self.tt_settings = tt_settings
+
     def __prepare_proxy_playwright(self, proxy: str) -> ProxySettings:
         splitted = urlsplit(proxy)
         if splitted.username and splitted.password:
@@ -411,7 +426,7 @@ class Capture():
             # Check if they are valid
             new_headers = {name.strip(): value.strip() for name, value in headers.items() if isinstance(name, str) and isinstance(value, str) and name.strip() and value.strip()}
             if new_headers != headers:
-                self.logger.warning(f'Headers contains invalid values:\n{json.dumps(headers, indent=2)}')
+                self.logger.warning(f'Headers contains invalid values:\n{orjson.dumps(headers, option=orjson.OPT_INDENT_2).decode()}')
         else:
             # This shouldn't happen, but we also cannot ensure the calls leading to this are following the specs,
             # and playwright dislikes invalid HTTP headers so we rather drop them.
@@ -1007,6 +1022,7 @@ class Capture():
                            with_screenshot: bool=True,
                            with_favicon: bool=False,
                            allow_tracking: bool=False,
+                           with_trusted_timestamps: bool=False,
                            ) -> CaptureResponse:
 
         to_return: CaptureResponse = {}
@@ -1199,6 +1215,11 @@ class Capture():
                                         rendered_hostname_only=rendered_hostname_only,
                                         max_depth_capture_time=max_capture_time,
                                         with_screenshot=with_screenshot)
+                                    if with_trusted_timestamps:
+                                        try:
+                                            await self._get_trusted_timestamps(child_capture)
+                                        except Exception as e:
+                                            self.logger.warning(f'Unable to get the trusted timestamps for the clild capture : {e}.')
                                     to_return['children'].append(child_capture)  # type: ignore[union-attr]
                             except (TimeoutError, asyncio.TimeoutError):
                                 self.logger.info(f'Timeout error, took more than {max_capture_time}s. Unable to capture {url}.')
@@ -1312,15 +1333,22 @@ class Capture():
                         self.logger.debug('Page closed.')
                         await self.context.close(reason="Closing the context because the capture finished.")  # context needs to be closed to generate the HAR
                         self.logger.debug('Context closed.')
-                        with open(self._temp_harfile.name) as _har:
-                            to_return['har'] = json.load(_har)
+                        with open(self._temp_harfile.name, 'rb') as _har:
+                            to_return['har'] = orjson.loads(_har.read())
                         self.logger.debug('Got HAR.')
+
                     if (to_return.get('har') and self.proxy and self.proxy.get('server')
                             and self.proxy['server'].startswith('socks5')):
                         # Only if the capture was not done via a socks5 proxy
                         if har := to_return['har']:  # Could be None
-                            async with timeout(30):
-                                await self.socks5_resolver(har)
+                            try:
+                                async with timeout(120):
+                                    await self.socks5_resolver(har)
+                            except (TimeoutError, asyncio.TimeoutError):
+                                self.logger.warning("Unable to resolve all the IPs via the socks5 proxy.")
+                                errors.append("Unable to resolve all the IPs via the socks5 proxy.")
+                                self.should_retry = True
+
                 except (TimeoutError, asyncio.TimeoutError):
                     self.logger.warning("Unable to close page and context at the end of the capture.")
                     errors.append("Unable to close page and context at the end of the capture.")
@@ -1331,8 +1359,62 @@ class Capture():
         self.logger.debug('Capture done')
         if errors:
             to_return['error'] = '\n'.join(errors)
+        if with_trusted_timestamps:
+            try:
+                await self._get_trusted_timestamps(to_return)
+            except Exception as e:
+                self.logger.warning(f'Unable to get trusted timestamps: {e}')
         return to_return
 
+    async def _get_trusted_timestamps(self, capture_response: CaptureResponse) -> None:
+        """Get trusted timestamps for the relevant values in the response"""
+        if not self.tt_settings:
+            self.logger.warning('The remote timestamper is not configured.')
+            return None
+        to_timestamp: dict[str, TimestampRequestBuilder] = {}
+        if last_redirected_url := capture_response.get('last_redirected_url'):
+            to_timestamp['last_redirected_url'] = TimestampRequestBuilder().data(last_redirected_url.encode())
+        if har := capture_response.get('har'):
+            to_timestamp['har'] = TimestampRequestBuilder().data(orjson.dumps(har))
+        if storage := capture_response.get('storage'):
+            to_timestamp['storage'] = TimestampRequestBuilder().data(orjson.dumps(storage))
+        if html := capture_response.get('html'):
+            to_timestamp['html'] = TimestampRequestBuilder().data(html.encode())
+        if png := capture_response.get('png'):
+            to_timestamp['png'] = TimestampRequestBuilder().data(png)
+        if downloaded_filename := capture_response.get('downloaded_filename'):
+            to_timestamp['downloaded_filename'] = TimestampRequestBuilder().data(downloaded_filename.encode())
+        if downloaded_file := capture_response.get('downloaded_file'):
+            to_timestamp['downloaded_file'] = TimestampRequestBuilder().data(downloaded_file)
+        # if potential_favicons := capture_response.get('potential_favicons'):
+        #    to_timestamp['potential_favicons'] = TimestampRequestBuilder().data(potential_favicons)
+
+        tt_requests: dict[str, TimeStampRequest] = {}
+        for k, trb in to_timestamp.items():
+            if h_algo := self.tt_settings.get('hash_algorithm'):
+                trb.hash_algorithm(h_algo)
+            tt_requests[k] = trb.build()
+
+        trusted_timestamps: dict[str, bytes] = {}
+
+        connector = None
+        if self.proxy and self.proxy.get('server'):
+            connector = ProxyConnector.from_url(self.proxy['server'])
+
+        timeout = aiohttp.ClientTimeout(total=10)
+        async with aiohttp.ClientSession(connector=connector, timeout=timeout) as session:
+            for k, tsr in tt_requests.items():
+                async with session.post(self.tt_settings['url'], data=tsr.as_bytes(),
+                                        headers={"Content-Type": "application/timestamp-query"}) as response:
+                    try:
+                        response.raise_for_status()
+                    except aiohttp.ClientResponseError as e:
+                        self.logger.warning(f'Unable to get Trusted Timestamp for {k}: {e}')
+                        continue
+                    trusted_timestamps[k] = await response.read()
+
+        capture_response['trusted_timestamps'] = {k: b64encode(v).decode() for k, v in trusted_timestamps.items()}
+
     async def _failsafe_get_screenshot(self, page: Page) -> bytes:
         self.logger.debug("Capturing a screenshot of the full page.")
         try:
@@ -1803,7 +1885,7 @@ class Capture():
     # END FAVICON EXTRACTOR
 
     # ##### Run DNS resolution over socks5 proxy #####
-    # This is only use when the capture is done over a socks5 proxy, and not on a .onion
+    # This is only use when the capture is done over a socks5 proxy, and not on a .onion / .i2p
     # We get the HAR file, iterate over the entries an update the IPs
 
     async def socks5_resolver(self, harfile: dict[str, Any]) -> None:
@@ -1814,10 +1896,12 @@ class Capture():
         for entry in harfile['log']['entries']:
             if entry['request']['url']:
                 parsed = urlparse(entry['request']['url'])
-                if parsed.netloc and not parsed.netloc.endswith('onion'):
-                    hostnames.add(parsed.netloc)
+                if (parsed.hostname
+                        and not parsed.hostname.endswith('onion')
+                        and not parsed.hostname.endswith('i2p')):
+                    hostnames.add(parsed.hostname)
         # use the same technique as in lookyloo to resolve many domains in parallel
-        semaphore = asyncio.Semaphore(20)
+        semaphore = asyncio.Semaphore(30)
         all_requests = [resolver.resolve(hostname, semaphore) for hostname in hostnames]
         await asyncio.gather(*all_requests)
         self.logger.debug('Resolved all domains through the proxy.')
